GraphBuilder = (function () {
    return function GraphBuilder (cAPI)
    {
        /*
         * Underscore stunt double
         */
        var _ = function(obj)
        {
            var o = Object.create(null);
            o.filter = function filter (func)
            {
                var filtered = [];
                for (var i = 0; i < obj.length; i++)
                {
                    if (func(obj[i]))
                    {
                        filtered.push(obj[i]);
                    }
                }
                return filtered;
            };
            
            o.map = function map (func)
            {
                var mapped = [];
                for (var i = 0; i < obj.length; i++)
                {
                    mapped.push(func(obj[i]));
                }
                return mapped;
            };
            
            o.each = function each (func)
            {
                for (var i = 0; i < obj.length; i++)
                {
                    func(obj[i]);
                }
            };
            
            return o;
        };

        this.Graph = function Graph (arg)
        {
            var name;
            
            /*
             * Private members
             */
            var graph = this;
            var structure;            
            
            if (this.constructor !== Graph)
            {
                return new Graph(arg);
            }
            
            if (typeof arg === "string")
            {
                name = arg;
            }
            else if (typeof arg === "object")
            {
                structure = arg;
                
                if ('name' in graph)
                {
                    name = structure.name;
                }
            }
            else if (typeof arg === "undefined")
            {
                name = "temp/temporary_graph_" + Math.round(Math.random() * 99999999);
            }
            
            /*
             * Public members
             */
            this.make = function ()
            {
                cAPI.graphBuilder.make(name);
                return this;
            };
            
            this.addNode = function (argument)
            {
                var pluginID, node;
                
                if(typeof argument === "object")
                {
                    var config = argument.config;
                    var parents = argument.parents;
                    pluginID = argument.pluginID;
                    node = this.addNode(pluginID).linkToParents(parents).configure(config);
                    markDirty();
                    return node;
                }
                else
                {
                    pluginID = argument;
                    node = new Node(cAPI.graphBuilder.createNode(name,pluginID).nodeID);
                    markDirty();
                    return node;
                }
            };

            this.node = function (nodeID)
            {
                return new Node(nodeID);
            };

            this.nodes = function ()
            {
                var nodes = [],
                    i = 0;
                
                if (arguments.length > 0 && typeof arguments[0] !== "function")
                {
                    for (i = 0; i < arguments.length; i++)
                    {
                        nodes.push(this.node(arguments[i]));
                    }
                    return nodes;
                }
                else
                {
                    var structure = parseGraph();
                    for (i = 0; i < structure.nodes.length; i++)
                    {
                        nodes.push(this.node(structure.nodes[i].nodeID));
                    }
                    
                    if (typeof arguments[0] === "function")
                    {
                        nodes = _(nodes).filter(arguments[0]);
                    }

                    return nodes;
                }
            };
            
            this.nodesOfType = function (type)
            {
                return this.nodes(function (node) {
                    return node.type().toLowerCase() === type;
                });
            };
            
            this.save = function ()
            {
                cAPI.graphBuilder.save(name);
                return this;
            };
            
            this.startCapture = function ()
            {
                cAPI.graphBuilder.startCapture(name);            
                return this;
            };

            this.asString = function ()
            {
                if (cAPI)
                {
                    return JSON.stringify(cAPI.graphBuilder.getGraph(name));
                }
                else if (structure !== undefined)
                {
                    return JSON.stringify(structure);
                }
            };

            this.name = function ()
            {
                return name;
            };
            
            function parseGraph ()
            {
                if (parseGraph.needsReparse || typeof structure === "undefined")
                {
                    structure = JSON.parse(graph.asString());
                    parseGraph.needsReparse = false;
                }
                return structure;
            }

            function markDirty ()
            {
                parseGraph.needsReparse = true;
            }
            
            function Node (id)
            {
                /*
                 * Public members
                 */
                this.linkParents = function (parents)
                {
                    var parentIDs = [],
                        i = 0,
                        j = 0;
                    
                    for (i = 0; i < parents.length; i++)
                    {
                        if (typeof parents[i] === "object" && parents[i].hasOwnProperty("id"))
                        {
                            parentIDs.push(parents[i].id());
                        }
                        else if (typeof parents[i] === "string")
                        {
                            /* jshint loopfunc:true */
                            var nodesForPlugin = this.graph().nodes(function(node) {
                                return node.pluginID() === parents[i];
                            });
                            /* jshint loopfunc:false */
                            
                            for (j = 0; j < nodesForPlugin.length; j++)
                            {
                                parentIDs.push(nodesForPlugin[j].id());
                            }
                        }
                        else if (typeof parents[i] === "function")
                        {
                            var nodesForFunction = this.graph().nodes(parents[i]);
                            for(j = 0; j < nodesForFunction.length; j++)
                            {
                                parentIDs.push(nodesForFunction[j].id());
                            }
                        }                    
                        else
                        {
                            parentIDs.push(parents[i]);
                        }
                    }
                    cAPI.graphBuilder.linkNode(name,id,parentIDs.join(","),"");
                    markDirty();
                    return this;
                };
                
                this.linkChildren = function (children)
                {
                    var childIDs = [],
                        i = 0,
                        j = 0;
                    
                    for (i = 0; i < children.length; i++)
                    {
                        if (typeof children[i] === "object" && children[i].hasOwnProperty("id"))
                        {
                            childIDs.push(children[i].id());
                        }
                        else if (typeof children[i] === "string")
                        {
                            /* jshint loopfunc:true */
                            var nodesForPlugin = this.graph().nodes(function(node) {
                                return node.pluginID() === children[i];
                            });
                            /* jshint loopfunc:false */
                            
                            for (j = 0; j < nodesForPlugin.length; j++)
                            {
                                childIDs.push(nodesForPlugin[j].id());
                            }
                        }
                        else if (typeof children[i] === "function")
                        {
                            var nodesForFunction = this.graph().nodes(children[i]);
                            for (j = 0; j < nodesForFunction.length; j++)
                            {
                                childIDs.push(nodesForFunction[j].id());
                            }
                        }
                        else
                        {
                            childIDs.push(children[i]);
                        }
                    }
                    cAPI.graphBuilder.linkNode(name,id,"",childIDs.join(","));
                    markDirty();
                    return this;
                };
                
                this.configure = function (configuration)
                {
                    cAPI.graphBuilder.configureNode(name,id,configuration);
                    markDirty();
                    return this;
                };
                
                this.remove = function ()
                {
                    cAPI.graphBuilder.removeNode(name,id);
                    markDirty();
                    return this;
                };
                
                this.invokeService = function (service, args, callback)
                {
                    if (!args)
                    {
                        args = {};
                    }
                    
                    if (callback)
                    {
                        cAPI.graphBuilder.invokeService(name,id,service,args,callback);
                        return this;
                    }
                    else
                    {
                        return cAPI.graphBuilder.invokeService(name,id,service,args);
                    }
                };
                
                this.parents = function (filter)
                {
                    var parents = _(_(parseGraph().nodes).filter(function(node) {
                        return node.connectedTo.indexOf(id) > -1;
                    })).map(function(node) {
                        return graph.node(node.nodeID);
                    });

                    if (typeof filter === "function")
                    {
                        parents = _(parents).filter(filter);
                    }

                    return parents;
                };
                
                this.children = function (filter)
                {
                    var children = _(parseNode().connectedTo).map(function(nodeID) {
                        return graph.node(nodeID);
                    });

                    if (typeof filter === "function")
                    {
                        children = _(children).filter(filter);
                    }

                    return children;
                };        
                
                this.ancestors = function (filter)
                {
                    var ancestors = [];
                    _(this.parents()).each(function (parent) {
                        ancestors.push(parent);
                        _(parent.ancestors()).each(function (ancestor) {
                            ancestors.push(ancestor);
                        });
                    });
                    
                    if (typeof filter === "function")
                    {
                        ancestors = _(ancestors).filter(filter);
                    }
                    
                    return ancestors;
                };
                
                this.descendants = function (filter)
                {
                    var descendants = [];
                    _(this.children()).each(function (child) {
                        descendants.push(child);
                        _(child.descendants()).each(function (descendant) {
                            descendants.push(descendant);
                        });
                    });
                    
                    if (typeof filter === "function")
                    {
                        descendants = _(descendants).filter(filter);
                    }
                    
                    return descendants;
                };
                
                this.configuration = function()
                {
                    return JSON.parse(parseNode().config);
                };
                
                this.config = this.configuration;
                
                this.options = function()
                {
                    return cAPI.graphBuilder.nodeOptions(name,id);
                };
                
                this.graph = function()
                {
                    return graph;
                };
                
                this.id = function()
                {
                    return id;
                };
                
                this.pluginID = function()
                {
                    return this.configuration().pluginID;
                };
                
                this.type = function()
                {
                    return this.configuration().type;
                };
                
                /*
                 * Private members
                 */
                function parseNode()
                {
                    return _(parseGraph().nodes).filter(function(node) {
                      return node.nodeID == id; 
                    })[0];
                }
            }
        };
    };
})();
